# Copyright (c) 2016-2017 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.


class OutcomeCache(object):
    """
    Base class for configuration outcome caching strategies. It does not
    implement a caching strategy itself (or, it implements the disable-cache
    strategy) but leaves the implementation of real strategies to subclasses.
    """

    def add(self, config, result):
        """
        Add a new configuration to the cache.

        :param config: The configuration to save.
        :param result: The outcome of the added configuration.
        """
        pass

    def lookup(self, config):
        """
        Cache lookup to find out the outcome of a given configuration.

        :param config: The configuration we are looking for.
        :return: PASS or FAIL if config is in the cache; None, otherwise.
        """
        return None

    def clear(self):
        """Clear the cache."""
        pass


class ConfigCache(OutcomeCache):

    class _Entry(object):
        """
        This class holds test outcomes for configurations. This avoids
        running the same test twice.

        The outcome cache is implemented as a tree.  Each node points
        to the outcome of the remaining list.

        Example: ([1, 2, 3], PASS), ([1, 2], FAIL), ([1, 4, 5], FAIL):

             (2, FAIL)--(3, PASS)
            /
        (1, None)
            \
             (4, None)--(5, FAIL)
        """

        def __init__(self):
            self.result = None  # Result so far
            self.tail = {}  # Points to outcome of tail

        def add(self, config, result):
            """
            Add (config, RESULT) to the cache. config must be a list of scalars.

            :param config: Config to add to the cache.
            :param result: The outcome of the added config.
            """
            p = self
            for cs in config:
                if cs not in p.tail:
                    p.tail[cs] = ConfigCache._Entry()
                p = p.tail[cs]
            p.result = result

        def lookup(self, config):
            """
            Performs a cache lookup.

            :param config: The configuration we are looking for.
            :return: PASS or FAIL if config is in the cache, None otherwise.
            """
            p = self
            for cs in config:
                if cs not in p.tail:
                    return None
                p = p.tail[cs]

            return p.result

        def _str(self, config=None):
            config = config or []
            s = ''
            if self.result is not None:
                s += '\t[%s]: %r,\n' % (', '.join([repr(c) for c in config]), self.result)
            for c, e in sorted(self.tail.items()):
                config.append(c)
                s += e._str(config)
                config.pop()
            return s

    def __init__(self):
        self._root = ConfigCache._Entry()

    def add(self, config, result):
        self._root.add(config, result)

    def lookup(self, config):
        return self._root.lookup(config)

    def clear(self):
        self._root = ConfigCache._Entry()

    def __str__(self):
        return '{\n%s}' % self._root._str()


class ContentCache(OutcomeCache):
    """
    Class that can cache the outcome of test cases by their content.
    """

    def __init__(self):
        self.container = dict()
        self.test_builder = None

    def set_test_builder(self, test_builder):
        self.test_builder = test_builder

    def add(self, config, result):
        self.container[self.test_builder(config)] = result

    def lookup(self, config):
        return self.container.get(self.test_builder(config), None)

    def clear(self):
        pass

    def __str__(self):
        from functools import reduce
        return '{\n%s}' % reduce(lambda x, y: x + y,
                                 ['\t%r: %r,\n' % (k, v)
                                  for k, v in sorted(self.container.items())],
                                 '')


# Aliases for cache classes to help their identification in CLI.
none = OutcomeCache
config = ConfigCache
content = ContentCache



# Copyright (c) 2016 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.


def zeller(config, n):
    """
    Splits up the input config into n pieces as used by Zeller in the original
    reference implementation. The approach works iteratively in n steps, first
    slicing off a chunk sized 1/n-th of the original config, then slicing off
    1/(n-1)-th of the remainder, and so on, until the last piece is halved
    (always using integers division).

    :param config: The configuration to split.
    :param n: The number of sets the configuration will be split up to.
    :return: List of the split sets.
    """
    subsets = []
    start = 0
    for i in range(n):
        subset = config[start:start + (len(config) - start) // (n - i)]
        subsets.append(subset)
        start += len(subset)
    return subsets


def balanced(config, n):
    """
    Slightly different version of Zeller's split. This version keeps the split
    balanced by distributing the residuals of the integer division among all
    chunks. This way, the size of the chunks in the resulting split is not monotonous.

    :param config: The configuration to split.
    :param n: The number of sets the configuration will be split up to.
    :return: List of the split sets.
    """
    return [config[len(config) * i // n: len(config) * (i + 1) // n] for i in range(n)]



# Copyright (c) 2018 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.


from logging import *


TRACE = DEBUG // 2
DISABLE = CRITICAL + 1


levels = {
    'TRACE': TRACE,
    'DEBUG': DEBUG,
    'INFO': INFO,
    'WARNING': WARNING,
    'ERROR': ERROR,
    'CRITICAL': CRITICAL,
    'DISABLE': DISABLE,
}


__getLogger = getLogger
def getLogger(name=None):
    logger = __getLogger(name)
    logger.trace = lambda msg, *args, **kwargs: logger.log(TRACE, msg, *args, **kwargs)
    return logger



# Copyright (c) 2016-2017 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import logging
import multiprocessing
import os
import psutil
import signal
import sys

logger = logging.getLogger(__name__)
is_windows = sys.platform.startswith('win32')


class Loop(object):
    """Parallel loop implementation based on multiprocessing module."""

    # TODO: could be None if we trusted that wait for infinity would not hang
    _timeout = 1

    def __init__(self, j=os.cpu_count(), max_utilization=100):
        """
        Initialize a parallel loop object.

        :param j: The maximum number of parallel jobs.
        :param max_utilization: The maximum CPU utilization. Above this no more new jobs
                                will be started.
        """
        self._j = j
        self._max_utilization = max_utilization
        # This gets initialized to 0, may be set to 1 anytime, but must not be reset to 0 ever;
        # thus, no locking is needed when accessing
        self._break = multiprocessing.sharedctypes.Value('i', 0, lock=False)
        self._lock = multiprocessing.Condition()
        self._slots = multiprocessing.sharedctypes.Array('i', j, lock=False)
        psutil.cpu_percent(None)

    # Beware! this is running in a new process now. state is shared with fork,
    # but only changes to shared objects will be visible in parent.
    def _body(self, i, target, args):
        """
        Executes the given function in its own process group (on Windows: in
        its own process).

        :param i: The index of the current configuration.
        :param target: The function to run in parallel.
        :param args: The arguments that the target should run with.
        """
        if not is_windows:
            os.setpgrp()

        try:
            if not target(*args):
                self._break.value = 1
        except:
            logger.warning('', exc_info=True)
            self._break.value = 1

        self._slots[i] = 0

        with self._lock:
            self._lock.notify()

    def _abort(self):
        """Terminate all live jobs."""
        for i, pid in enumerate(self._slots):
            if pid != 0:
                if not is_windows:
                    try:
                        os.killpg(pid, signal.SIGTERM)
                    except OSError:
                        # If the process with pid did not have time to become a process group leader,
                        # then pgid does not exist and os.killpg could not kill the process,
                        # so re-try kill the process only.
                        try:
                            os.kill(pid, signal.SIGTERM)
                        except OSError:
                            pass
                else:
                    root_proc = psutil.Process(pid)
                    children = root_proc.children(recursive=True) + [root_proc]
                    for proc in children:
                        try:
                            # Would be easier to use proc.terminate() here but psutils
                            # (up to version 5.4.0) on Windows terminates processes with
                            # the 0 signal/code, making the outcome of the terminated
                            # process indistinguishable from a successful execution.
                            os.kill(proc.pid, signal.SIGTERM)
                        except OSError:
                            pass
                    psutil.wait_procs(children, timeout=1)
            self._slots[i] = 0

    def _cleanup_slots(self):
        for i, pid in enumerate(self._slots):
            if pid != 0:
                if not psutil.pid_exists(pid):
                    self._slots[i] = 0

    # Target is expected to return True if loop shall continue and False if it
    # should break.
    def do(self, target, args):
        """
        Execute the target function if there is empty slot for it and
        the CPU utilization is suitable.

        :param target: The function to run in parallel.
        :param args: The parameters to run the target function with.
        :return: False if an interesting configuration was found. Otherwise returns True.
        """
        if self._break.value:
            logger.debug('do() called on a broken loop')
            self._abort()
            return False

        i = None
        while True:
            if psutil.cpu_percent(None) <= self._max_utilization:
                self._cleanup_slots()
                i = next((x for x in range(self._j) if self._slots[x] == 0), None)
                if i is not None:
                    break
            with self._lock:
                self._lock.wait(self._timeout)

        if self._break.value:
            self._abort()
            return False

        proc = multiprocessing.Process(target=self._body, args=(i, target, args))
        proc.start()
        self._slots[i] = proc.pid

        return True

    def join(self):
        """Wait until all the parallel child processes have finished."""
        if self._break.value:
            self._abort()
            return

        while True:
            self._cleanup_slots()
            # Return if all the jobs are done, that's there isn't any non-zero value in self._slots.
            if not any(self._slots):
                return

            with self._lock:
                self._lock.wait(self._timeout)

            if self._break.value:
                self._abort()



# Copyright (c) 2016-2018 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import logging

from .outcome_cache import OutcomeCache

logger = logging.getLogger(__name__)


class AbstractDD(object):
    """Abstract super-class of the parallel and non-parallel DD classes."""

    # Test outcomes.
    PASS = 'PASS'
    FAIL = 'FAIL'

    def __init__(self, test, split, *, cache=None, id_prefix=()):
        """
        Initialise an abstract DD class. Not to be called directly,
        only by super calls in subclass initializers.

        :param test: A callable tester object.
        :param split: Splitter method to break a configuration up to n parts.
        :param cache: Cache object to use.
        :param id_prefix: Tuple to prepend to config IDs during tests.
        """
        self._test = test
        self._split = split
        self._cache = cache or OutcomeCache()
        self._id_prefix = id_prefix

    def ddmin(self, config, *, n=2):
        """
        Return a 1-minimal failing subset of the initial configuration.

        :param config: The initial configuration that will be reduced.
        :param n: The number of sets that the config is initially split to.
        :return: 1-minimal failing configuration.
        """
        if len(config) < 2:
            assert self._test_config(config, ('assert',)) == self.FAIL
            logger.info('Test case is minimal already.')
            return config

        run = 1
        n = min(len(config), n)
        complement_offset = 0

        while True:
            assert self._test_config(config, ('r%d' % run, 'assert')) == self.FAIL

            subsets = self._split(config, n)

            logger.info('Run #%d: trying %s.', run, ' + '.join([str(len(subsets[i])) for i in range(n)]))

            next_config, next_n, complement_offset = self._reduce_config(run, config, subsets, complement_offset)

            if next_config is None:
                # Minimization ends if no interesting configuration was found by the finest splitting.
                if n == len(config):
                    logger.info('Done.')
                    return config

                next_config = config
                next_n = min(len(config), n * 2)
                complement_offset = (complement_offset * next_n) / n
                logger.info('Increase granularity to %d.', next_n)

            else:
                # Interesting configuration is found.
                logger.info('Reduced to %d units.', len(next_config))
                logger.debug('New config: %r.', next_config)

                # Minimization ends if the configuration is already reduced to a single unit.
                if len(next_config) == 1:
                    logger.info('Done.')
                    return next_config

            config = next_config
            n = next_n
            run += 1

    def _reduce_config(self, run, config, subsets, complement_offset):
        """
        Perform the reduce task of ddmin. To be overridden by subclasses.

        :param run: The index of the current iteration.
        :param config: The current configuration under testing.
        :param subsets: List of sets that the current configuration is split to.
        :param complement_offset: A compensation offset needed to calculate the index
               of the first unchecked complement (optimization purpose only).
        :return: Tuple: (failing config or None, next n or None, next complement_offset).
        """
        pass

    def _lookup_cache(self, config, config_id):
        """
        Perform a cache lookup if caching is enabled.

        :param config: The configuration we are looking for.
        :param config_id: The ID describing the configuration (only for debug message).
        :return: None if outcome is not found for config in cache or if caching is disabled, PASS or FAIL otherwise.
        """
        cached_result = self._cache.lookup(config)
        if cached_result is not None:
            logger.debug('\t[ %s ]: cache = %r', self._pretty_config_id(self._id_prefix + config_id), cached_result)

        return cached_result

    def _test_config(self, config, config_id):
        """
        Test a single configuration and save the result in cache.

        :param config: The current configuration to test.
        :param config_id: Unique ID that will be used to save tests to easily identifiable directories.
        :return: PASS or FAIL
        """
        config_id = self._id_prefix + config_id

        logger.debug('\t[ %s ]: test...', self._pretty_config_id(config_id))
        outcome = self._test(config, config_id)
        logger.debug('\t[ %s ]: test = %r', self._pretty_config_id(config_id), outcome)

        if 'assert' not in config_id:
            self._cache.add(config, outcome)

        return outcome

    @staticmethod
    def _pretty_config_id(config_id):
        """
        Create beautified identifier for the current task from the argument.
        The argument is typically a tuple in the form of ('rN', 'DM'), where N
        is the index of the current iteration, D is direction of reduce (either
        s(ubset) or c(omplement)), and M is the index of the current test in the
        iteration. Alternatively, argument can also be in the form of
        (rN, 'assert') for double checking the input at the start of an
        iteration.

        :param config_id: Config ID tuple.
        :return: Concatenating the arguments with slashes, e.g., "rN / DM".
        """
        return ' / '.join(str(i) for i in config_id)

    @staticmethod
    def _minus(c1, c2):
        """Return a list of all elements of C1 that are not in C2."""
        c2 = set(c2)
        return [c for c in c1 if c not in c2]



# Copyright (c) 2016 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

forward = range  #: Generator returning numbers from 0 to n-1.


def backward(n):
    """
    Generator returning numbers from n - 1 to 0 decreasing by 1.

    :param n: Upper bound of the interval.
    :return: Decreasing numbers from n - 1 to 0.
    """
    yield from range(n - 1, -1, -1)


def skip(n):
    """
    Do not return anything. Used to skip subset (or, less often, complement) checks.

    :param n: Anything. It won't ever be used. It's added only for consistency reasons.
    :return: None
    """
    yield from ()


def random(n):
    """
    Returns numbers 0..n-1 in random order.

    :param n: Upper bound of the interval.
    :return: Numbers in random order from 0 to n - 1.
    """
    from random import shuffle

    lst = list(range(n))
    shuffle(lst)
    yield from lst



# Copyright (c) 2016-2017 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import inspect

from multiprocessing import Lock
from multiprocessing.managers import BaseManager

shared_cache_class_store = dict()


class SharedDataManager(BaseManager):
    """Data manager to share the cache object between parallel processes."""
    pass


class SharedCacheTrampoline(object):
    """Thread-safe wrapper for SharedCache methods."""

    def __init__(self, shared_cache, fn):
        self._shared_cache = shared_cache
        self._fn = fn

    def __call__(self, *args, **kwargs):
        with self._shared_cache._lock:
            return getattr(self._shared_cache._cache, self._fn)(*args, **kwargs)


class SharedCache(object):
    """Thread-safe cache representation that stores the evaluated configurations and their outcome."""

    def __init__(self, cache):
        self._cache = cache
        self._lock = Lock()

        for fn, _ in inspect.getmembers(cache.__class__, lambda m: inspect.isfunction(m) and m.__name__ != '__init__'):
            setattr(self, fn, SharedCacheTrampoline(self, fn))


class SharedCacheConstructor(object):
    """Wrapper for cache classes to instantiate them as shared."""

    def __init__(self, cache_class):
        self._cache_class = cache_class

    def __call__(self, *args, **kwargs):
        return SharedCache(self._cache_class(*args, **kwargs))


def shared_cache_decorator(cache_class):
    global shared_cache_class_store

    if cache_class not in shared_cache_class_store:
        SharedDataManager.register(cache_class.__name__, SharedCacheConstructor(cache_class), None,
                                   [fn for fn, _ in inspect.getmembers(cache_class, lambda m: inspect.isfunction(m) and m.__name__ != '__init__')])
        getattr(SharedDataManager, cache_class.__name__).__signature__ = inspect.signature(cache_class)

        data_manager = SharedDataManager()
        data_manager.start()

        shared_cache_class_store[cache_class] = getattr(data_manager, cache_class.__name__)

    return shared_cache_class_store[cache_class]



# Copyright (c) 2016-2018 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import logging

from . import config_iterators
from . import config_splitters
from .abstract_dd import AbstractDD
from .outcome_cache import ConfigCache

logger = logging.getLogger(__name__)


class LightDD(AbstractDD):
    """Single process version of the Delta Debugging algorithm."""

    def __init__(self, test, *, cache=None, id_prefix=(), split=config_splitters.zeller,
                 subset_first=True, subset_iterator=config_iterators.forward, complement_iterator=config_iterators.forward):
        """
        Initialize a LightDD object.

        :param test: A callable tester object.
        :param cache: Cache object to use.
        :param id_prefix: Tuple to prepend to config IDs during tests.
        :param split: Splitter method to break a configuration up to n parts.
        :param subset_first: Boolean value denoting whether the reduce has to start with the subset-based approach or not.
        :param subset_iterator: Reference to a generator function that provides config indices in an arbitrary order.
        :param complement_iterator: Reference to a generator function that provides config indices in an arbitrary order.
        """
        cache = cache or ConfigCache()
        AbstractDD.__init__(self, test, split, cache=cache, id_prefix=id_prefix)

        self._subset_first = subset_first
        self._subset_iterator = subset_iterator
        self._complement_iterator = complement_iterator

    def _reduce_config(self, run, config, subsets, complement_offset):
        """
        Perform the reduce task in single process mode.

        :param run: The index of the current iteration.
        :param config: The current configuration under testing.
        :param subsets: List of sets that the current configuration is split to.
        :param complement_offset: A compensation offset needed to calculate the index
               of the first unchecked complement (optimization purpose only).
        :return: Tuple: (failing config or None, next n or None, next complement_offset).
        """
        if self._subset_first:
            first_reduce, second_reduce = self._reduce_to_subset, self._reduce_to_complement
        else:
            first_reduce, second_reduce = self._reduce_to_complement, self._reduce_to_subset

        next_config, next_n, complement_offset = first_reduce(run, config, subsets, complement_offset)
        if next_config is None:
            next_config, next_n, complement_offset = second_reduce(run, config, subsets, complement_offset)

        return next_config, next_n, complement_offset

    def _reduce_to_subset(self, run, config, subsets, complement_offset):
        """
        Perform a subset-based reduce task.

        :param run: The index of the current iteration.
        :param config: The current configuration under testing.
        :param subsets: List of sets that the current configuration is split to.
        :param complement_offset: A compensation offset needed to calculate the index
               of the first unchecked complement (optimization purpose only).
        :return: Tuple: (failing config or None, next n or None, next complement_offset).
        """
        n = len(subsets)
        for i in self._subset_iterator(n):
            if i is None:
                continue

            config_id = ('r%d' % run, 's%d' % i)

            # Get the outcome either from cache or by testing it.
            outcome = self._lookup_cache(subsets[i], config_id) or self._test_config(subsets[i], config_id)
            if outcome == self.FAIL:
                # Interesting subset is found.
                return subsets[i], 2, 0

        return None, None, complement_offset

    def _reduce_to_complement(self, run, config, subsets, complement_offset):
        """
        Perform a complement-based reduce task.

        :param run: The index of the current iteration.
        :param config: The current configuration under testing.
        :param subsets: List of sets that the current configuration is split to.
        :param complement_offset: A compensation offset needed to calculate the index
               of the first unchecked complement (optimization purpose only).
        :return: Tuple: (failing config or None, next n or None, next complement_offset).
        """
        n = len(subsets)
        for i in self._complement_iterator(n):
            if i is None:
                continue
            i = int((i + complement_offset) % n)

            config_id = ('r%d' % run, 'c%d' % i)
            complement = self._minus(config, subsets[i])

            outcome = self._lookup_cache(complement, config_id) or self._test_config(complement, config_id)
            if outcome == self.FAIL:
                # Interesting complement is found.
                # In next run, start removing the following subset
                return complement, max(n - 1, 2), i

        return None, None, complement_offset



# Copyright (c) 2016-2018 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

from .abstract_dd import AbstractDD
from .abstract_parallel_dd import AbstractParallelDD
from .cli import __version__, call
from .combined_iterator import CombinedIterator
from .combined_parallel_dd import CombinedParallelDD
from .light_dd import LightDD
from .outcome_cache import ConfigCache, ContentCache, OutcomeCache
from .parallel_dd import ParallelDD
from .shared_cache import shared_cache_decorator
from .subprocess_test import ConcatTestBuilder, SubprocessTest


__all__ = [
    '__version__',
    'AbstractDD',
    'AbstractParallelDD',
    'call',
    'cli',
    'CombinedIterator',
    'CombinedParallelDD',
    'ConcatTestBuilder',
    'ConfigCache',
    'config_iterators',
    'config_splitters',
    'ContentCache',
    'LightDD',
    'logging',
    'OutcomeCache',
    'ParallelDD',
    'shared_cache_decorator',
    'SubprocessTest',
]



# Copyright (c) 2016-2018 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import argparse
import chardet
import codecs
import os
import pkgutil
import time

from os.path import abspath, basename, exists, join, relpath
from shutil import rmtree

from . import config_splitters, config_iterators, outcome_cache
from . import logging
from .combined_iterator import CombinedIterator
from .combined_parallel_dd import CombinedParallelDD
from .light_dd import LightDD
from .parallel_dd import ParallelDD
from .shared_cache import shared_cache_decorator
from .subprocess_test import ConcatTestBuilder, SubprocessTest

logger = logging.getLogger('picire')
__version__ = pkgutil.get_data(__package__, 'VERSION').decode('ascii').strip()


def create_parser():
    def int_or_inf(value):
        if value == 'inf':
            return float('inf')
        value = int(value)
        if value < 2:
            raise argparse.ArgumentTypeError('invalid value: {value!r} (must be at least 2)'.format(value=value))
        return value

    parser = argparse.ArgumentParser(description='Command line interface of the "picire" test case reducer')
    parser.add_argument('-i', '--input', metavar='FILE', required=True,
                        help='test case to be reduced')

    # Base reduce settings.
    parser.add_argument('--cache', metavar='NAME',
                        choices=[i for i in dir(outcome_cache) if not i.startswith('_') and i.islower()], default='config',
                        help='cache strategy (%(choices)s; default: %(default)s)')
    parser.add_argument('--split', metavar='NAME', dest='split_method',
                        choices=[i for i in dir(config_splitters) if not i.startswith('_')], default='zeller',
                        help='split algorithm (%(choices)s; default: %(default)s)')
    parser.add_argument('--test', metavar='FILE', required=True,
                        help='test command that decides about interestingness of an input')
    parser.add_argument('--granularity', type=int_or_inf, default=2,
                        help='initial splitting granularity (integer or \'inf\'; default: %(default)d).')
    parser.add_argument('--encoding', metavar='NAME',
                        help='test case encoding (default: autodetect)')

    # Extra settings for parallel reduce.
    parser.add_argument('-p', '--parallel', action='store_true', default=False,
                        help='run DD in parallel')
    parser.add_argument('-c', '--combine-loops', action='store_true', default=False,
                        help='combine subset and complement check loops for more parallelization (has effect in parallel mode only)')
    parser.add_argument('-j', '--jobs', metavar='N', type=int, default=os.cpu_count(),
                        help='maximum number of test commands to execute in parallel (has effect in parallel mode only; default: %(default)s)')
    parser.add_argument('-u', '--max-utilization', metavar='N', type=int, default=100,
                        help='maximum CPU utilization allowed; don\'t start new parallel jobs until utilization is higher (has effect in parallel mode only; default: %(default)s)')

    # Tweaks how to walk through the chunk lists.
    parser.add_argument('--complement-first', dest='subset_first', action='store_false', default=True,
                        help='check complements first')
    parser.add_argument('--subset-iterator', metavar='NAME',
                        choices=[i for i in dir(config_iterators) if not i.startswith('_')], default='forward',
                        help='ordering strategy for looping through subsets (%(choices)s; default: %(default)s)')
    parser.add_argument('--complement-iterator', metavar='NAME',
                        choices=[i for i in dir(config_iterators) if not i.startswith('_')], default='forward',
                        help='ordering strategy for looping through complements (%(choices)s; default: %(default)s)')

    # Additional settings.
    parser.add_argument('-l', '--log-level', metavar='LEVEL',
                        choices=sorted(logging.levels.keys(), key=lambda k: logging.levels[k]), default='INFO',
                        help='verbosity level of diagnostic messages (%(choices)s; default: %(default)s)')
    parser.add_argument('-v', '--verbose', dest='log_level', action='store_const', const='DEBUG', default=argparse.SUPPRESS,
                        help='verbose mode (alias for -l %(const)s)')
    parser.add_argument('-q', '--quiet', dest='log_level', action='store_const', const='DISABLE', default=argparse.SUPPRESS,
                        help='quiet mode (alias for -l %(const)s)')
    parser.add_argument('-o', '--out', metavar='DIR',
                        help='working directory (default: input.timestamp)')
    parser.add_argument('--disable-cleanup', dest='cleanup', default=True, action='store_false',
                        help='disable the removal of generated temporary files')
    return parser


def process_args(parser, args):
    args.log_level = logging.levels[args.log_level]

    args.input = abspath(relpath(args.input))
    if not exists(args.input):
        parser.error('Input file does not exits: %s' % args.input)

    with open(args.input, 'rb') as f:
        args.src = f.read()

    if args.encoding:
        try:
            codecs.lookup(args.encoding)
        except LookupError:
            parser.error('The given encoding (%s) is not known.' % args.encoding)
    else:
        args.encoding = chardet.detect(args.src)['encoding']
        if not args.encoding:
            parser.error('The encoding of the test is not recognized.'
                         'Please define it with the --encoding command line option.')

    args.test = abspath(relpath(args.test))
    if not exists(args.test) or not os.access(args.test, os.X_OK):
        parser.error('Tester program does not exist or isn\'t executable: %s' % args.test)

    args.tester_class = SubprocessTest
    args.tester_config = {
        'encoding': args.encoding,
        'command_pattern': '%s %%s' % args.test
    }

    args.cache = getattr(outcome_cache, args.cache)
    if args.parallel:
        args.cache = shared_cache_decorator(args.cache)

    split_method = getattr(config_splitters, args.split_method)
    subset_iterator = getattr(config_iterators, args.subset_iterator)
    complement_iterator = getattr(config_iterators, args.complement_iterator)

    # Choose the reducer class that will be used and its configuration.
    args.reduce_config = {'split': split_method}
    if not args.parallel:
        args.reduce_class = LightDD
        args.reduce_config['subset_iterator'] = subset_iterator
        args.reduce_config['complement_iterator'] = complement_iterator
        args.reduce_config['subset_first'] = args.subset_first
    else:
        args.reduce_config['proc_num'] = args.jobs
        args.reduce_config['max_utilization'] = args.max_utilization

        if args.combine_loops:
            args.reduce_class = CombinedParallelDD
            args.reduce_config['config_iterator'] = CombinedIterator(args.subset_first,
                                                                     subset_iterator,
                                                                     complement_iterator)
        else:
            args.reduce_class = ParallelDD
            args.reduce_config['subset_iterator'] = subset_iterator
            args.reduce_config['complement_iterator'] = complement_iterator
            args.reduce_config['subset_first'] = args.subset_first

    args.out = abspath(relpath(args.out if args.out else '%s.%s' % (args.input, time.strftime('%Y%m%d_%H%M%S'))))


def call(*,
         reduce_class, reduce_config,
         tester_class, tester_config,
         input, src, encoding, out,
         atom='line', granularity=2,
         cache_class=None, cleanup=True):
    """
    Execute picire as if invoked from command line, however, control its
    behaviour not via command line arguments but function parameters.

    :param reduce_class: Reference to the reducer class.
    :param reduce_config: Dictionary containing information to initialize the reduce_class.
    :param tester_class: Reference to a runnable class that can decide about the interestingness of a test case.
    :param tester_config: Dictionary containing information to initialize the tester_class.
    :param input: Path to the test case to reduce (only used to determine the name of the output file).
    :param src: Contents of the test case to reduce.
    :param encoding: Encoding of the input test case.
    :param out: Path to the output directory.
    :param atom: Input granularity to work with during reduce ('char', 'line', or 'both'; default: 'line').
    :param granularity: Initial granularity (default: 2).
    :param cache_class: Reference to the cache class to use.
    :param cleanup: Binary flag denoting whether removing auxiliary files at the end is enabled (default: True).
    :return: The path to the minimal test case.
    """

    # Get the parameters in a dictionary so that they can be pretty-printed
    # (minus src, as that parameter can be arbitrarily large)
    args = locals().copy()
    del args['src']
    logger.info('Reduce session starts for %s\n%s',
                input, ''.join(['\t%s: %s\n' % (k, v) for k, v in sorted(args.items())]))

    tests_dir = join(out, 'tests')
    # Split source to the chosen atoms.
    if atom in ['line', 'both']:
        content = src.decode(encoding).splitlines(keepends=True)
        tests_dir = join(tests_dir, 'line')
    elif atom == 'char':
        content = src.decode(encoding)
        tests_dir = join(tests_dir, 'char')
    os.makedirs(tests_dir, exist_ok=True)
    logger.info('Initial test contains %d %ss', len(content), atom)

    test_builder = ConcatTestBuilder(content)
    cache = cache_class() if cache_class else None
    if hasattr(cache, 'set_test_builder'):
        cache.set_test_builder(test_builder)

    dd = reduce_class(tester_class(test_builder=test_builder,
                                   test_pattern=join(tests_dir, '%s', basename(input)),
                                   **tester_config),
                      cache=cache,
                      **reduce_config)
    min_set = dd.ddmin(list(range(len(content))), n=granularity)

    logger.trace('The cached results are: %s', cache)
    logger.debug('A minimal config is: %r', min_set)

    out_file = join(out, basename(input))
    out_src = test_builder(min_set)
    with codecs.open(out_file, 'w', encoding=encoding, errors='ignore') as f:
        f.write(out_src)
    logger.info('Result is saved to %s.', out_file)

    if cleanup:
        rmtree(tests_dir)

    if atom == 'both':
        out_file = call(reduce_class=reduce_class, reduce_config=reduce_config,
                        tester_class=tester_class, tester_config=tester_config,
                        input=out_file, src=out_src.encode(encoding=encoding), encoding=encoding, out=out,
                        atom='char',
                        cache_class=cache_class, cleanup=cleanup)

    return out_file


def execute():
    """
    The main entry point of picire.
    """

    parser = create_parser()
    # Implementation specific CLI options that are not needed to be part of the core parser.
    parser.add_argument('-a', '--atom', metavar='NAME', choices=['char', 'line', 'both'], default='line',
                        help='atom (i.e., granularity) of input (%(choices)s; default: %(default)s)')
    parser.add_argument('--version', action='version', version='%(prog)s {version}'.format(version=__version__))

    args = parser.parse_args()
    process_args(parser, args)

    logging.basicConfig(format='%(message)s')
    logger.setLevel(args.log_level)

    call(reduce_class=args.reduce_class,
         reduce_config=args.reduce_config,
         tester_class=args.tester_class,
         tester_config=args.tester_config,
         input=args.input,
         src=args.src,
         encoding=args.encoding,
         out=args.out,
         atom=args.atom,
         granularity=args.granularity,
         cache_class=args.cache,
         cleanup=args.cleanup)



# Copyright (c) 2016-2018 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import codecs
import os
import shlex
import sys

from subprocess import Popen

from .abstract_dd import AbstractDD


class SubprocessTest(object):

    def __init__(self, command_pattern, test_builder, test_pattern, encoding):
        """
        Wrapper around the script provided by the user. It decides about the
        interestingness based on the return code of executed script.

        :param command_pattern: The command as the tester script should be ran. Except
                                that the path of the test is substituted with %s.
        :param test_builder: Callable object that creates test case from a configuration.
        :param test_pattern: The patter of the test's path. It contains one %s part
                             that will be replaced with the ID of the certain configurations.
        :param encoding: The encoding that will be used to save the tests.
        """
        self.command_pattern = command_pattern
        self.encoding = encoding
        self.test_builder = test_builder
        self.test_pattern = test_pattern

    def __call__(self, config, config_id):
        """
        Saving and evaluating of the current configuration.

        :param config: The list of units (chars or lines) that have to
                       be compiled into a single test.
        :param config_id: Unique ID of the current configuration. It's used to
                          name the containing folder of the current test.
        :return: The evaluation of the current test. It's either FAIL or PASS.
        """

        test_path = self.test_pattern % '_'.join(str(i) for i in config_id)
        test_dir = os.path.dirname(test_path)

        os.makedirs(test_dir, exist_ok=True)

        with codecs.open(test_path, 'w', encoding=self.encoding, errors='ignore') as f:
            f.write(self.test_builder(config))

        with Popen(shlex.split(self.command_pattern % test_path,
                               posix=not sys.platform.startswith('win32')),
                   cwd=test_dir) as proc:
            proc.wait()

        # Determine outcome.
        if proc.returncode == 0:
            return AbstractDD.FAIL
        else:
            return AbstractDD.PASS


class ConcatTestBuilder(object):
    """Callable class that builds test case from a configuration."""

    def __init__(self, content):
        """
        Initialize a test builder with the atoms (e.g. chars or lines) of the original test case.

        :param content: Atoms of the original test case.
        """
        self._content = content

    def __call__(self, config):
        """
        Builds test case from the given config.

        :param config: Configuration to build a test case from.
        :return: Test case described by the config.
        """
        return ''.join([self._content[x] for x in config])
